use regex::Regex;
use serde::{Serialize, Deserialize};
use std::{error::Error, fs, collections::HashMap, path::Path};

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct ResourceCollection(HashMap<String, Resource>);

impl ResourceCollection {
    pub fn validate(self: &Self) -> Result<(), String> {
        for (name, resource) in self.0.iter() {
            resource.validate(Some(name.clone()))?;
        }
        Ok(())
    }

    pub fn init(self: &mut Self) -> Result<(), String> {
        for (name, resource) in self.0.iter_mut() {
            resource.init(Some(name.clone()))?;
        }
        Ok(())
    }

    pub fn get<T: Into<String>>(self: &Self, key: T) -> Option<&Resource> {
        self.0.get(&key.into())
    }
}

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct Resource {
    pub path: String,
    pub mime: String,
    cached_data: Option<Vec<u8>>
}

impl Resource {

    pub fn init(self: &mut Self, name: Option<String>) -> Result<(), String> {
        if self.cached_data.is_none() {
            self.read_to_cache().map_err(|x|
                if let Some(name) = name {
                    format!("Error while initalizing resource {name}: {:?}", x)
                } else {
                    format!("Error while initalizing unknown resource: {:?}", x)
                }
            )?
        }
        Ok(())
    }

    pub fn validate(self: &Self, name: Option<String>) -> Result<(), String> {
        // dont catch this panic
        let re = Regex::new(r"^[a-z][a-z0-9_]*(\.[a-z0-9_]+)+[0-9a-z_]$").unwrap();

        if let Some(name) = name.clone() {
            if ! re.is_match(name.as_str()) {
                return Err(format!("Invalid resource name: {name}"));
            }
        }

        let path = Path::new(&self.path);
        if ! path.is_file() {
            if let Some(name) = name {
                return Err(format!("Resource {name} is not a file!"));
            } else {
                return Err(format!("A resource with path {} is not a file!", self.path));
            }
        }
        Ok(())
    }

    fn read_to_cache(self: &mut Self) -> Result<(), Box<dyn Error>> {
        let data = fs::read(self.path.clone())?;
        self.cached_data = Some(data);
        Ok(())
    }

    pub fn get(self: &Self) -> Result<Vec<u8>, Box<dyn Error>> {
        if let Some(cached) = self.cached_data.clone() {
            Ok(cached)
        } else {
            Ok(fs::read(self.path.clone())?)
        }
    }
}